/****************************************************************************
 * This file is part of the project AqFinance.
 * AqFinance (c) by 2018 Martin Preuss, all rights reserved.
 *
 * The license for this file can be found in the file COPYING which you
 * should have received along with this file.
 ****************************************************************************/


#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include "ff_editstatements.hpp"
#include "ff_app.hpp"
#include "ff_editpartnerstmrule.hpp"

#include <aqfinance/engine/book/ae_query_stm.h>

#include <gwenhywfar/i18n.h>
#include <gwenhywfar/debug.h>
#include <gwenhywfar/url.h>

#include <gwen-gui-fox16/theme.h>
#include <gwen-gui-fox16/fox16_htmllabel.hpp>


#define I18N(msg) GWEN_I18N_Translate(PACKAGE, msg)



FXDEFMAP(FF_EditStatements) FF_EditStatementsMap[]={
  FXMAPFUNC(SEL_COMMAND,            FXDialogBox::ID_ACCEPT,                  FF_EditStatements::onCmdAccept),
  FXMAPFUNC(SEL_COMMAND,            FXDialogBox::ID_CANCEL,                  FF_EditStatements::onCmdCancel),
  FXMAPFUNC(SEL_RIGHTBUTTONRELEASE, FF_EditStatements::ID_STATEMENT_LIST,    FF_EditStatements::onListRightBtnReleased),
  FXMAPFUNC(SEL_COMMAND,            FF_EditStatements::ID_FILTER_BY_PURPOSE, FF_EditStatements::onCmdFilterByPurpose),
  FXMAPFUNC(SEL_COMMAND,            FF_EditStatements::ID_FILTER_BY_ACCOUNT, FF_EditStatements::onCmdFilterByAccount),
  FXMAPFUNC(SEL_COMMAND,            FF_EditStatements::ID_FILTER_BY_NAME,    FF_EditStatements::onCmdFilterByName),
  FXMAPFUNC(SEL_COMMAND,            FF_EditStatements::ID_FILTER_BY_IBAN,    FF_EditStatements::onCmdFilterByIban),
  FXMAPFUNC(SEL_COMMAND,            FF_EditStatements::ID_AUTOMATCH,         FF_EditStatements::onCmdAutoMatch),
  FXMAPFUNC(SEL_REPLACED,           FF_EditStatements::ID_STATEMENT_LIST,    FF_EditStatements::onTableContentReplaced),
};


FXIMPLEMENT(FF_EditStatements, FXDialogBox,
            FF_EditStatementsMap, ARRAYNUMBER(FF_EditStatementsMap))




FF_EditStatements::FF_EditStatements(FF_App* a,
				     FXComposite *parent,
				     const FXString& title,
				     FXuint opts,
				     FXint x, FXint y, FXint w, FXint h,
				     FXint pl, FXint pr, FXint pt, FXint pb,
				     FXint hs, FXint vs)
:FXDialogBox(parent, title, opts, x, y, w, h, pl, pr, pt, pb, hs, vs)
,m_app(a)
,m_label(NULL)
,m_statementTable(NULL) {
  FXVerticalFrame *vf;
  FXHorizontalFrame *hf;

  vf=new FXVerticalFrame(this, LAYOUT_FILL_X | LAYOUT_FILL_Y);
  m_label=new FOX16_HtmlLabel(vf,
			      I18N("<p>You can now edit the statements below.</p>"
				   "<p>Most importantly, you can assign <b>business partners</b> "
				   "to the account statements below.</p>"
				   "<p>Double-click into a field or begin typing to start editing</p>"),
			      LAYOUT_FILL_X);
  m_label->setMinimumWidth(150);

  m_statementTable=new FF_StatementTable(vf, this, ID_STATEMENT_LIST,
					 FXXTable::TABLE_TRACK_SCROLLBARS |
					 LAYOUT_FILL_X | LAYOUT_FILL_Y);
  m_statementTable->addColumn(FF_StatementTable::COLUMN_TYPE_DATE, 100);
  m_statementTable->addColumn(FF_StatementTable::COLUMN_TYPE_MEMO, 100);
  m_statementTable->addColumn(FF_StatementTable::COLUMN_TYPE_PARTNER, 100);
  m_statementTable->addColumn(FF_StatementTable::COLUMN_TYPE_VALUE, 100);

  m_statementTable->setMinRowHeight(20);
  m_statementTable->setColorizeMatches(true);

  new FXHorizontalSeparator(vf, LAYOUT_FILL_X | SEPARATOR_GROOVE);
  hf=new FXHorizontalFrame(vf, LAYOUT_FILL_X | PACK_UNIFORM_WIDTH);

  new THEMEBUTTON(hf, I18N("Automatch"), NULL, this, ID_AUTOMATCH,
		  BUTTON_NORMAL);
  new FXSpring(hf, LAYOUT_FILL_X);

  new THEMEBUTTON(hf, I18N("Ok"), NULL, this, ID_ACCEPT,
		  BUTTON_NORMAL | BUTTON_DEFAULT | BUTTON_INITIAL);
  new THEMEBUTTON(hf, I18N("Cancel"), NULL, this, ID_CANCEL,
		  BUTTON_NORMAL);
}



FF_EditStatements::FF_EditStatements()
:FXDialogBox()
,m_app(NULL)
,m_label(NULL)
,m_statementTable(NULL) {
}



FF_EditStatements::~FF_EditStatements() {
}



void FF_EditStatements::init() {
  GWEN_DB_NODE *db;
  int w, h;
  int i;
  int sortByColumn;
  int sortUp;

  db=m_app->getDialogConfig("FF_EditStatements");
  if (db==NULL)
    db=GWEN_DB_Group_new("config");

  w=GWEN_DB_GetIntValue(db, "width", 0, 0);
  h=GWEN_DB_GetIntValue(db, "height", 0, 0);
  if (w>0 && h>0)
    resize(w, h);
  else
    resize(getWidth(), getDefaultHeight()+100);

  sortByColumn=GWEN_DB_GetIntValue(db, "sortByColumn", 0, -1);
  sortUp=GWEN_DB_GetIntValue(db, "sortUp", 0, 0);

  for (i=0; i<m_statementTable->getNumHeaders(); i++) {
    int j=GWEN_DB_GetIntValue(db, "columns", i, m_statementTable->getHeaderSize(i));
    m_statementTable->setHeaderSize(i, j);
    if (i==sortByColumn) {
      if (sortUp)
        m_statementTable->setHeaderArrowDir(i, TRUE);
      else
        m_statementTable->setHeaderArrowDir(i, FALSE);
    }
  }

  GWEN_DB_Group_free(db);
}



void FF_EditStatements::fini() {
  GWEN_DB_NODE *db;
  int rv;
  int i;

  db=GWEN_DB_Group_new("config");

  GWEN_DB_SetIntValue(db, GWEN_DB_FLAGS_OVERWRITE_VARS,
                      "width", getWidth());
  GWEN_DB_SetIntValue(db, GWEN_DB_FLAGS_OVERWRITE_VARS,
                      "height", getHeight());

  for (i=0; i<m_statementTable->getNumHeaders(); i++) {
    FXbool b;

    GWEN_DB_SetIntValue(db, GWEN_DB_FLAGS_DEFAULT, "columns",
                        m_statementTable->getHeaderSize(i));
    b=m_statementTable->getHeaderArrowDir(i);
    if (b!=MAYBE) {
      GWEN_DB_SetIntValue(db, GWEN_DB_FLAGS_OVERWRITE_VARS, "sortByColumn", i);
      if (b==TRUE)
        GWEN_DB_SetIntValue(db, GWEN_DB_FLAGS_OVERWRITE_VARS, "sortUp", 1);
      else
        GWEN_DB_SetIntValue(db, GWEN_DB_FLAGS_OVERWRITE_VARS, "sortUp", 0);
    }
  }

  rv=m_app->setDialogConfig("FF_EditStatements", db);
  if (rv<0) {
    DBG_ERROR(0, "here (%d)", rv);
  }
  GWEN_DB_Group_free(db);
}



FXint FF_EditStatements::checkGui(FXbool quiet) {
  return 0;
}



void FF_EditStatements::toGui(const AE_STATEMENT_LIST *tlist) {
  AE_STATEMENT *t;

  m_statementTable->reload();
  t=AE_Statement_List_First(tlist);
  while(t) {
    AE_STATEMENT *cpy;

    cpy=AE_Statement_dup(t);
    m_statementTable->addStatement(cpy);
    t=AE_Statement_List_Next(t);
  }

  autoMatch();
  m_statementTable->sort();
}



void FF_EditStatements::fromGui(AE_STATEMENT_LIST *tlist) {
  int i;
  int cnt;

  cnt=m_statementTable->getNumberOfRows();
  for (i=0; i<cnt; i++) {
    AE_STATEMENT *t;

    t=m_statementTable->getStatementByRow(i);
    if (t) {
      /* add copy to list because the statement table owns the list elements */
      AE_Statement_List_Add(AE_Statement_dup(t), tlist);
    }
  }
}



void FF_EditStatements::create() {
  FXDialogBox::create();
  m_label->create();
  m_statementTable->create();
}



long FF_EditStatements::onCmdAccept(FXObject*, FXSelector, void *ptr) {
  DBG_ERROR(0, "Accept");
  if (checkGui(FALSE)==0) {
    getApp()->stopModal(this,TRUE);
    hide();
  }
  return 1;
}



long FF_EditStatements::onCmdCancel(FXObject*, FXSelector, void *ptr) {
  DBG_ERROR(0, "Cancel");
  getApp()->stopModal(this,FALSE);
  hide();
  return 1;
}



long FF_EditStatements::onListRightBtnReleased(FXObject*, FXSelector, void *ptr) {
  FXEvent* event=(FXEvent*)ptr;

  if (!event->moved){
    FXMenuPane pmenu(this);
    FXMenuPane paneFilter(this);

    new FXMenuCommand(&paneFilter, I18N("by purpose"), NULL, this, ID_FILTER_BY_PURPOSE);
    new FXMenuCommand(&paneFilter, I18N("by name"), NULL, this, ID_FILTER_BY_NAME);
    new FXMenuSeparator(&paneFilter);
    new FXMenuCommand(&paneFilter, I18N("by IBAN"), NULL, this, ID_FILTER_BY_IBAN);
    new FXMenuCommand(&paneFilter, I18N("by account"), NULL, this, ID_FILTER_BY_ACCOUNT);

    new FXMenuCascade(&pmenu, I18N("Create a Filter Rule"), NULL, &paneFilter);
    new FXMenuSeparator(&pmenu);
    new FXMenuCommand(&pmenu, I18N("Back"), NULL, &pmenu, ID_UNPOST);

    pmenu.create();
    pmenu.popup(NULL, event->root_x, event->root_y);
    getApp()->runModalWhileShown(&pmenu);
  }

  return 1;
}



long FF_EditStatements::onCmdFilterByPurpose(FXObject*, FXSelector, void *ptr) {
  AE_STATEMENT *st;
  AE_RULE *r;
  GWEN_DB_NODE *dbExample=NULL;

  DBG_ERROR(0, "Filter by Purpose");

  r=AE_Rule_new();
  AE_Rule_SetOwnerType(r, AE_Book_TableType_Partner);
  AE_Rule_SetTypeId(r, AE_Book_TableType_BankStatement);

  st=m_statementTable->getCurrentStatement();
  if (st) {
    const char *s;

    s=AE_Statement_GetPurpose(st);
    if (s && *s) {
      GWEN_BUFFER *tbuf;

      tbuf=GWEN_Buffer_new(0, 256, 0, 1);
      GWEN_Buffer_AppendString(tbuf, "purpose\t?contains\t");
      GWEN_Buffer_AppendString(tbuf, s);
      GWEN_Buffer_AppendString(tbuf, "\n");
      AE_Rule_SetExpression(r, GWEN_Buffer_GetStart(tbuf));
      GWEN_Buffer_free(tbuf);
    }

    dbExample=GWEN_DB_Group_new("example");
    AE_Statement_toDb(st, dbExample);
  }

  if (FF_EditPartnerStatementRule::editRule(m_app, this,
					    I18N("Create a Filter Rule"),
					    r,
                                            dbExample)) {
    int rv;

    rv=AE_Book_AddRule(m_app->getBook(), AE_Book_TableType_PartnerRule, r, 1);
    if (rv<0) {
      DBG_ERROR(0, "here (%d)", rv);
    }
    else {
      // TODO: change partner
      autoMatch();
    }
  }

  m_statementTable->reload();

  GWEN_DB_Group_free(dbExample);
  AE_Rule_free(r);

  return 1;
}



long FF_EditStatements::onCmdFilterByAccount(FXObject*, FXSelector, void *ptr) {
  AE_STATEMENT *st;
  AE_RULE *r;
  GWEN_DB_NODE *dbExample=NULL;

  DBG_ERROR(0, "Filter by Account");

  r=AE_Rule_new();
  AE_Rule_SetOwnerType(r, AE_Book_TableType_Partner);
  AE_Rule_SetTypeId(r, AE_Book_TableType_BankStatement);
  AE_Rule_SetCombineMode(r, AE_RuleCombineMode_And);

  st=m_statementTable->getCurrentStatement();
  if (st) {
    const char *s;
    GWEN_BUFFER *tbuf;

    tbuf=GWEN_Buffer_new(0, 256, 0, 1);

    s=AE_Statement_GetRemoteBankCode(st);
    if (s && *s) {
      GWEN_Buffer_AppendString(tbuf, "remotebankcode\t==\t");
      GWEN_Buffer_AppendString(tbuf, s);
      GWEN_Buffer_AppendString(tbuf, "\n");
    }

    s=AE_Statement_GetRemoteAccountNumber(st);
    if (s && *s) {
      GWEN_Buffer_AppendString(tbuf, "remoteaccountnumber\t==\t");
      GWEN_Buffer_AppendString(tbuf, s);
      GWEN_Buffer_AppendString(tbuf, "\n");
    }

    if (GWEN_Buffer_GetUsedBytes(tbuf))
      AE_Rule_SetExpression(r, GWEN_Buffer_GetStart(tbuf));
    GWEN_Buffer_free(tbuf);

    dbExample=GWEN_DB_Group_new("example");
    AE_Statement_toDb(st, dbExample);
  }

  if (FF_EditPartnerStatementRule::editRule(m_app, this,
					    I18N("Create a Filter Rule"),
					    r,
					    dbExample)) {
    int rv;

    rv=AE_Book_AddRule(m_app->getBook(), AE_Book_TableType_PartnerRule, r, 1);
    if (rv<0) {
      DBG_ERROR(0, "here (%d)", rv);
    }
    else {
      // TODO: change account
      autoMatch();
    }
  }
  m_statementTable->reload();

  GWEN_DB_Group_free(dbExample);
  AE_Rule_free(r);

  return 1;
}



long FF_EditStatements::onCmdFilterByIban(FXObject*, FXSelector, void *ptr) {
  AE_STATEMENT *st;
  AE_RULE *r;
  GWEN_DB_NODE *dbExample=NULL;

  DBG_ERROR(0, "Filter by Account");

  r=AE_Rule_new();
  AE_Rule_SetOwnerType(r, AE_Book_TableType_Partner);
  AE_Rule_SetTypeId(r, AE_Book_TableType_BankStatement);
  AE_Rule_SetCombineMode(r, AE_RuleCombineMode_And);

  st=m_statementTable->getCurrentStatement();
  if (st) {
    const char *s;
    GWEN_BUFFER *tbuf;

    tbuf=GWEN_Buffer_new(0, 256, 0, 1);

    s=AE_Statement_GetRemoteBic(st);
    if (s && *s) {
      GWEN_Buffer_AppendString(tbuf, "remotebic\t==\t");
      GWEN_Buffer_AppendString(tbuf, s);
      GWEN_Buffer_AppendString(tbuf, "\n");
    }

    s=AE_Statement_GetRemoteIban(st);
    if (s && *s) {
      GWEN_Buffer_AppendString(tbuf, "remoteiban\t==\t");
      GWEN_Buffer_AppendString(tbuf, s);
      GWEN_Buffer_AppendString(tbuf, "\n");
    }

    if (GWEN_Buffer_GetUsedBytes(tbuf))
      AE_Rule_SetExpression(r, GWEN_Buffer_GetStart(tbuf));
    GWEN_Buffer_free(tbuf);

    dbExample=GWEN_DB_Group_new("example");
    AE_Statement_toDb(st, dbExample);
  }

  if (FF_EditPartnerStatementRule::editRule(m_app, this,
					    I18N("Create a Filter Rule"),
					    r,
					    dbExample)) {
    int rv;

    rv=AE_Book_AddRule(m_app->getBook(), AE_Book_TableType_PartnerRule, r, 1);
    if (rv<0) {
      DBG_ERROR(0, "here (%d)", rv);
    }
    else {
      // TODO: change account
      autoMatch();
    }
  }
  m_statementTable->reload();

  GWEN_DB_Group_free(dbExample);
  AE_Rule_free(r);

  return 1;
}



long FF_EditStatements::onCmdFilterByName(FXObject*, FXSelector, void *ptr) {
  AE_STATEMENT *st;
  AE_RULE *r;
  GWEN_DB_NODE *dbExample=NULL;

  DBG_ERROR(0, "Filter by Name");

  r=AE_Rule_new();
  AE_Rule_SetOwnerType(r, AE_Book_TableType_Partner);
  AE_Rule_SetTypeId(r, AE_Book_TableType_BankStatement);

  st=m_statementTable->getCurrentStatement();
  if (st) {
    const char *s;

    s=AE_Statement_GetRemoteName(st);
    if (s && *s) {
      GWEN_BUFFER *tbuf;

      tbuf=GWEN_Buffer_new(0, 256, 0, 1);
      GWEN_Buffer_AppendString(tbuf, "remotename\t?contains\t");
      GWEN_Buffer_AppendString(tbuf, s);
      GWEN_Buffer_AppendString(tbuf, "\n");
      AE_Rule_SetExpression(r, GWEN_Buffer_GetStart(tbuf));
      GWEN_Buffer_free(tbuf);
    }

    dbExample=GWEN_DB_Group_new("example");
    AE_Statement_toDb(st, dbExample);
  }

  if (FF_EditPartnerStatementRule::editRule(m_app, this,
					    I18N("Create a Filter Rule"),
					    r,
					    dbExample)) {
    int rv;

    rv=AE_Book_AddRule(m_app->getBook(), AE_Book_TableType_PartnerRule, r, 1);
    if (rv<0) {
      DBG_ERROR(0, "here (%d)", rv);
    }
    else {
      // TODO: change account
      autoMatch();
    }
  }
  m_statementTable->reload();

  GWEN_DB_Group_free(dbExample);
  AE_Rule_free(r);

  return 1;
}



long FF_EditStatements::onCmdAutoMatch(FXObject*, FXSelector, void *ptr) {
  autoMatch();
  return 1;
}


void FF_EditStatements::autoMatch() {
  int cnt;

  cnt=m_statementTable->getNumberOfRows();
  if (cnt) {
    AE_RULE_LIST *rules;
    int rv;
    char qbuf[64];

    snprintf(qbuf, sizeof(qbuf)-1, "$ownerType==%d", AE_Book_TableType_Partner);
    qbuf[sizeof(qbuf)-1]=0;

    rules=AE_Rule_List_new();
    rv=AE_Book_QueryRuleList(m_app->getBook(), AE_Book_TableType_PartnerRule, 0, qbuf, rules);
    if (rv<0) {
      DBG_ERROR(0, "here (%d)", rv);
      AE_Rule_List_free(rules);
    }
    else {
      if (AE_Rule_List_GetCount(rules)==0) {
        DBG_ERROR(0, "No rules");
      }
      else {
	int i;
	AE_RULE *r;
	GWEN_BUFFER *tbuf;
	bool needRebuild=false;

	/* create queries for rules and store them */
	tbuf=GWEN_Buffer_new(0, 256, 0, 1);
        r=AE_Rule_List_First(rules);
	while(r) {
	  rv=AE_Rule_toQueryString(r, tbuf);
	  if (rv<0) {
	    DBG_ERROR(0, "here (%d)", rv);
	  }
	  else {
	    AQDB_QUERY *q;

	    q=AE_StatementQuery_new(GWEN_Buffer_GetStart(tbuf));
	    if (q) {
	      DBG_ERROR(0, "Created query for [%s]", GWEN_Buffer_GetStart(tbuf));
	      AE_Rule_SetQuery(r, q);
	    }
	    else {
	      DBG_ERROR(0, "Bad query string [%s]", GWEN_Buffer_GetStart(tbuf));
	    }
	  }
	  GWEN_Buffer_Reset(tbuf);
	  r=AE_Rule_List_Next(r);
	}
	GWEN_Buffer_free(tbuf);

	/* check statements */
	for (i=0; i<cnt; i++) {
	  AE_STATEMENT *st;

	  st=m_statementTable->getStatementByRow(i);
	  if (st) {
	    uint32_t flags;

	    flags=AE_Statement_GetRuntimeFlags(st);
	    /* don't modify manually set statements */
	    if ((flags & AE_STATEMENT_RTFLAGS_MATCHMASK)!=AE_STATEMENT_RTFLAGS_MATCHMANUAL) {
	      AE_RULE *r;
	      int matches=0;
	      AQDB_ID matchedPid=0;

	      r=AE_Rule_List_First(rules);
	      while(r) {
		AQDB_QUERY *q;
		AQDB_ID pid;

		q=AE_Rule_GetQuery(r);
		pid=AE_Rule_GetOwnerId(r);
		if (q && pid) {
		  rv=AE_StatementQuery_Check(q, st);
		  if (rv>0) {
		    if (matchedPid==0) {
		      matchedPid=pid;
		      matches++;
		    }
		    else {
		      if (matchedPid!=pid)
			matches++;
		    }
		  }
		  else if (rv<0) {
		    DBG_ERROR(0, "here (%d)", rv);
		  }
		}

		r=AE_Rule_List_Next(r);
	      }

	      if (matches) {
		DBG_ERROR(0, "%d matches", matches);

		/* clear match mask */
		flags&=~AE_STATEMENT_RTFLAGS_MATCHMASK;
		if (matches==1)
		  /* only one rule matches, so we are quite sure abouth the match */
		  flags|=AE_STATEMENT_RTFLAGS_MATCHAUTO2;
		else
		  /* at least two rules matched, so we can't be sure about the result */
		  flags|=AE_STATEMENT_RTFLAGS_MATCHAUTO1;
		AE_Statement_SetRuntimeFlags(st, flags);
		AE_Statement_SetPartnerId(st, matchedPid);
		AE_Statement_AddRuntimeFlags(st, AE_STATEMENT_RTFLAGS_MODIFIED);
		needRebuild=true;
	      }
	    }
	  }
	}
	if (needRebuild)
	  m_statementTable->rebuild();
	AE_Rule_List_free(rules);
      }
    }
  }
  else {
    DBG_ERROR(0, "No statements");
  }
}



long FF_EditStatements::onTableContentReplaced(FXObject*, FXSelector, void* ptr) {
  FXXTable::TableCoords *pos=(FXXTable::TableCoords*) ptr;
  int r;
  AE_STATEMENT *st;

  r=pos->row;

  st=m_statementTable->getStatementByRow(r);
  if (st) {
    AE_Statement_SubRuntimeFlags(st, AE_STATEMENT_RTFLAGS_MATCHMASK);
    AE_Statement_AddRuntimeFlags(st, AE_STATEMENT_RTFLAGS_MATCHMANUAL);
    m_statementTable->updateRow(r);
  }

  return 1;
}



bool FF_EditStatements::editStatements(FF_App* app,
				       FXComposite *parent,
				       const FXString &title,
				       AE_STATEMENT_LIST *tlist) {
  FF_EditStatements dlg(app, parent, title, DECOR_ALL);
  FXuint dlgres;

  dlg.resize(440, 400);
  dlg.create();
  dlg.init();
  dlg.toGui(tlist);

  dlg.show(PLACEMENT_VISIBLE);
  dlg.layout();

  dlgres=dlg.execute(PLACEMENT_VISIBLE);
  dlg.fini();
  DBG_ERROR(0, "Result: %d", dlgres);
  if (dlgres==0) {
    return false;
  }

  AE_Statement_List_Clear(tlist);
  dlg.fromGui(tlist);
  return true;
}



