/****************************************************************************
 * This file is part of the project AqFinance.
 * AqFinance (c) by 2010 Martin Preuss, all rights reserved.
 *
 * The license for this file can be found in the file COPYING which you
 * should have received along with this file.
 ****************************************************************************/


#ifdef HAVE_CONFIG_H
# include <config.h>
#endif


#include "fm_businesspage.hpp"
#include "fm_invoicepage.hpp"
#include "ff_invoicetree.hpp"
#include "ff_app.hpp"
#include "ff_editjob.hpp"

#include <aqfinance/engine/modules/ae_openbills.h>


#include <gwenhywfar/debug.h>
#include <gwenhywfar/i18n.h>
#include <gwenhywfar/gwendate.h>


#define I18N(msg) GWEN_I18N_Translate(PACKAGE, msg)



FXDEFMAP(FM_InvoicePage) FM_InvoicePageMap[]={
  FXMAPFUNC(SEL_COMMAND, FM_InvoicePage::ID_BUTTON_NEW_INVOICE,
            FM_InvoicePage::onCmdNewInvoice),
  FXMAPFUNC(SEL_COMMAND, FM_InvoicePage::ID_BUTTON_ADD_PAYMENT,
            FM_InvoicePage::onCmdAddPayment),

  FXMAPFUNC(SEL_CHANGED, FM_InvoicePage::ID_INVOICE_TREE,
            FM_InvoicePage::onTreeChanged),
  FXMAPFUNC(SEL_RIGHTBUTTONPRESS, FM_InvoicePage::ID_INVOICE_TREE,
            FM_InvoicePage::onTreeRightBtnPress),
};



FXIMPLEMENT(FM_InvoicePage, FXHorizontalFrame,
            FM_InvoicePageMap, ARRAYNUMBER(FM_InvoicePageMap))






FM_InvoicePage::FM_InvoicePage(bool isIncoming,
			       FM_BusinessPage *mainPage,
			       FXComposite *p,
			       FXuint opts,
			       FXint x, FXint y, FXint w, FXint h)
:FXHorizontalFrame(p, opts, x, y, w, h)
,m_isIncoming(isIncoming)
,m_mainPage(mainPage)
,m_needRebuild(true)
,m_isVisible(false)
{
  FXVerticalFrame *vf;
  FXIcon *ic;
  FF_App *app;

  flags|=FLAG_ENABLED|FLAG_SHOWN;
  app=dynamic_cast<FF_App*>(getApp());

  m_invoiceTree=new FF_InvoiceTree(this, this, ID_INVOICE_TREE,
				   LAYOUT_SIDE_LEFT |
				   LAYOUT_FILL_X | LAYOUT_FILL_Y |
				   FOLDINGLIST_BROWSESELECT |
				   FOLDINGLIST_SHOWS_LINES |
				   FOLDINGLIST_SHOWS_BOXES |
				   FOLDINGLIST_ROOT_BOXES |
				   FRAME_SUNKEN|FRAME_THICK);
  ic=app->getIcon("minifolderopen");
  if (ic)
    m_invoiceTree->setIconOpen(ic);
  ic=app->getIcon("minifolder");
  if (ic)
    m_invoiceTree->setIconClosed(ic);

  vf=new FXVerticalFrame(this, LAYOUT_FILL_Y | PACK_UNIFORM_WIDTH);

  m_newInvoiceButton=new FXButton(vf,
				  I18N("New Invoice"),
				  NULL,
				  this, ID_BUTTON_NEW_INVOICE,
				  BUTTON_NORMAL);
  m_addPaymentButton=new FXButton(vf,
				  I18N("Add Payment"),
				  NULL,
				  this, ID_BUTTON_ADD_PAYMENT,
				  BUTTON_NORMAL);

}




FM_InvoicePage::FM_InvoicePage()
:FXHorizontalFrame()
,m_isIncoming(true)
,m_mainPage(NULL)
,m_invoiceTree(NULL)
,m_needRebuild(false)
{
}



FM_InvoicePage::~FM_InvoicePage() {
}



void FM_InvoicePage::init(GWEN_DB_NODE *db) {
  GWEN_DB_NODE *dbT;

  DBG_INFO(0, "Init");
  dbT=GWEN_DB_GetGroup(db, GWEN_PATH_FLAGS_NAMEMUSTEXIST, "invoicePage");
  if (dbT)
    dbT=GWEN_DB_GetGroup(dbT, GWEN_PATH_FLAGS_NAMEMUSTEXIST,
			 m_isIncoming?"incoming":"outgoing");
  if (dbT) {
    int i;
    int sortByColumn;
    int sortUp;

    sortByColumn=GWEN_DB_GetIntValue(dbT, "sortByColumn", 0, -1);
    sortUp=GWEN_DB_GetIntValue(dbT, "sortUp", 0, 0);

    for (i=0; i<m_invoiceTree->getNumHeaders(); i++) {
      int j=GWEN_DB_GetIntValue(dbT, "columns", i, m_invoiceTree->getHeaderSize(i));
      m_invoiceTree->setHeaderSize(i, j);
      if (i==sortByColumn) {
        if (sortUp)
          m_invoiceTree->setHeaderArrowDir(i, TRUE);
        else
          m_invoiceTree->setHeaderArrowDir(i, FALSE);
      }
    }
  }
}



void FM_InvoicePage::fini(GWEN_DB_NODE *db) {
  GWEN_DB_NODE *dbT;
  int i;

  DBG_INFO(0, "Fini");
  dbT=GWEN_DB_GetGroup(db, GWEN_DB_FLAGS_DEFAULT, "invoicePage");
  dbT=GWEN_DB_GetGroup(dbT, GWEN_DB_FLAGS_OVERWRITE_GROUPS, m_isIncoming?"incoming":"outgoing");
  for (i=0; i<m_invoiceTree->getNumHeaders(); i++) {
    FXbool b;

    GWEN_DB_SetIntValue(dbT, GWEN_DB_FLAGS_DEFAULT, "columns",
                        m_invoiceTree->getHeaderSize(i));
    b=m_invoiceTree->getHeaderArrowDir(i);
    if (b!=MAYBE) {
      GWEN_DB_SetIntValue(dbT, GWEN_DB_FLAGS_OVERWRITE_VARS, "sortByColumn", i);
      if (b==TRUE)
        GWEN_DB_SetIntValue(dbT, GWEN_DB_FLAGS_OVERWRITE_VARS, "sortUp", 1);
      else
        GWEN_DB_SetIntValue(dbT, GWEN_DB_FLAGS_OVERWRITE_VARS, "sortUp", 0);
    }
  }
}


long FM_InvoicePage::onCmdNewInvoice(FXObject*, FXSelector, void*) {
  FF_App *app;
  AE_BOOK *b;

  app=dynamic_cast<FF_App*>(getApp());
  b=app->getBook();
  if (b) {
    AE_JOB *j;
    GWEN_DATE *dt;

    j=AE_Job_new();
    dt=GWEN_Date_CurrentDate();
    AE_Job_SetEnterDate(j, dt);
    AE_Job_SetType(j, AE_Job_TypeInvoice);
  
    if (FF_EditJob::editJob(app, this,
			    I18N("Create New Invoice"),
			    j,
			    m_isIncoming /* incoming means: buy */)) {
      int rv;
      AE_TASK *tk;


      DBG_ERROR(0, "About to store job");
      tk=AE_Task_new();
      if (m_isIncoming)
	AE_Task_AddFlags(tk, AE_TASK_FLAGS_INCOMING);
      AE_Task_AddFlags(tk, AE_TASK_FLAGS_ACTIVE);

      AE_Task_SetEnterDate(tk, dt);
      AE_Task_SetWaitingFor(tk, AE_Job_TypePayment);

      rv=AE_Book_BeginEdit(b, 0);
      if (rv<0) {
	DBG_ERROR(0, "Error on begin edit: %d", rv);
	AE_Task_free(tk);
        GWEN_Date_free(dt);
	AE_Job_free(j);
	return 1;
      }

      DBG_ERROR(0, "Adding task");
      rv=AE_Book_AddTask(b, AE_Book_TableType_Task, tk, 0);
      if (rv<0) {
	DBG_ERROR(0, "here (%d)", rv);
	AE_Book_EndEdit(b, AQDB_ACTION_FLAGS_ABORT);
	AE_Task_free(tk);
        GWEN_Date_free(dt);
	AE_Job_free(j);
	return 1;
      }

      DBG_ERROR(0, "Applying job");
      AE_Job_SetTaskId(j, AE_Task_GetId(tk));
      rv=AE_Book_ApplyJob(b, j, NULL, 0);
      if (rv<0) {
	DBG_ERROR(0, "here (%d)", rv);
	AE_Book_EndEdit(b, AQDB_ACTION_FLAGS_ABORT);
	AE_Task_free(tk);
        GWEN_Date_free(dt);
	AE_Job_free(j);
	return 1;
      }

      rv=AE_Book_EndEdit(b, 0);
      if (rv<0) {
	DBG_ERROR(0, "here (%d)", rv);
	AE_Book_EndEdit(b, AQDB_ACTION_FLAGS_ABORT);
	AE_Task_free(tk);
	GWEN_Date_free(dt);
	AE_Job_free(j);
	return 1;
      }
      AE_Task_free(tk);
    }
    GWEN_Date_free(dt);
    AE_Job_free(j);
  }

  return 1;
}



long FM_InvoicePage::onCmdAddPayment(FXObject*, FXSelector, void*) {
  FF_App *app;
  AE_BOOK *b;

  app=dynamic_cast<FF_App*>(getApp());
  b=app->getBook();
  if (b) {
    AE_JOB *j;

    j=m_invoiceTree->getCurrentInvoice();
    if (j) {
      int rv;
      AE_JOB *jPayment;
      GWEN_DATE *dt;

      rv=AE_Book_ReadJobSubObjects(b, j, m_isIncoming);
      if (rv<0) {
	DBG_ERROR(0, "Error reading Job sub-objects: %d", rv);
        return 1;
      }

      dt=GWEN_Date_CurrentDate();

      jPayment=AE_Job_dup(j);
      AE_Job_SetReferenceId(jPayment, AE_Job_GetId(j));
      AE_Job_SetType(jPayment, AE_Job_TypePayment);
      AE_Job_SetId(jPayment, 0);
      AE_Job_SetDbId(jPayment, 0);

      AE_Job_SetEnterDate(jPayment, dt);
      AE_Job_SetDueDate(jPayment, dt);
      AE_Job_SetType(j, AE_Job_TypeInvoice);

      if (FF_EditJob::editJob(app, this,
			      I18N("Create Payment for Invoice"),
			      jPayment,
			      m_isIncoming /* incoming means: buy */)) {
	AE_TASK *tk=NULL;
	AQDB_ID tid;

	DBG_ERROR(0, "Paid");

	tid=AE_Job_GetTaskId(j);

	rv=AE_Book_BeginEdit(b, 0);
	if (rv<0) {
	  DBG_ERROR(0, "Error on begin edit: %d", rv);
	  AE_Task_free(tk);
	  GWEN_Date_free(dt);
	  AE_Job_free(j);
	  return 1;
	}

	/* read task */
	if (tid) {
	  tk=AE_Task_new();
	  rv=AE_Book_ReadTask(b, AE_Book_TableType_Task, tid, &tk);
	  if (rv<0) {
	    AE_Task_free(tk);
	    GWEN_Date_free(dt);
	    AE_Job_free(jPayment);
	    return 1;
	  }
	}

	/* modify task */
	AE_Task_SetWaitingFor(tk, AE_Job_TypeNone);

	DBG_ERROR(0, "Writing task");
	rv=AE_Book_WriteTask(b, AE_Book_TableType_Task, tk, 0);
	if (rv<0) {
	  DBG_ERROR(0, "here (%d)", rv);
	  AE_Book_EndEdit(b, AQDB_ACTION_FLAGS_ABORT);
	  AE_Task_free(tk);
	  GWEN_Date_free(dt);
	  AE_Job_free(j);
	  return 1;
	}
  
	DBG_ERROR(0, "Applying job");
	AE_Job_SetTaskId(jPayment, AE_Task_GetId(tk));
	rv=AE_Book_ApplyJob(b, jPayment, NULL, 0);
	if (rv<0) {
	  DBG_ERROR(0, "here (%d)", rv);
	  AE_Book_EndEdit(b, AQDB_ACTION_FLAGS_ABORT);
	  AE_Task_free(tk);
	  GWEN_Date_free(dt);
	  AE_Job_free(j);
	  return 1;
	}
  
	rv=AE_Book_EndEdit(b, 0);
	if (rv<0) {
	  DBG_ERROR(0, "here (%d)", rv);
	  AE_Book_EndEdit(b, AQDB_ACTION_FLAGS_ABORT);
	  AE_Task_free(tk);
	  GWEN_Date_free(dt);
	  AE_Job_free(j);
	  return 1;
	}
	AE_Task_free(tk);

        dbReload(AE_Book_TableType_Job);
      }

      AE_Job_free(jPayment);
      GWEN_Date_free(dt);
    }
  }

  return 1;
}



long FM_InvoicePage::onTreeChanged(FXObject*, FXSelector, void* ptr) {
  if (ptr) {
    FF_GroupTreeItem *item=dynamic_cast<FF_GroupTreeItem*>((FXFoldingItem*) ptr);
    if (item) {
      if (item->isDataItem()) {
        m_addPaymentButton->enable();
      }
      else {
	m_addPaymentButton->disable();
      }
    }
    else {
      m_addPaymentButton->disable();
    }
  }
  return 1;
}



long FM_InvoicePage::onTreeRightBtnPress(FXObject*, FXSelector, void* ptr) {
#if 0
  FXEvent* event=(FXEvent*)ptr;
  if (!event->moved){
    FXMenuPane pmenu(this);
    FXMenuCommand *c;
    AE_GROUP *gr;
    AE_STATEMENT *st;

    gr=m_invoiceTree->getCurrentGroup();
    st=m_invoiceTree->getCurrentTransfer();

    new FXMenuCommand(&pmenu, I18N("Add Group"), NULL, this, ID_BUTTON_ADDGROUP);
    new FXMenuCommand(&pmenu, I18N("Add Transfer"), NULL, this, ID_BUTTON_ADDTRANSFER);

    new FXMenuSeparator(&pmenu);

    c=new FXMenuCommand(&pmenu, I18N("Edit Group"), NULL, this, ID_BUTTON_EDITGROUP);
    if (!gr) c->disable();

    c=new FXMenuCommand(&pmenu, I18N("Edit Group"), NULL, this, ID_BUTTON_EDITGROUP);
    if (!gr) c->disable();

    c=new FXMenuCommand(&pmenu, I18N("Delete Group"), NULL, this, ID_BUTTON_DELGROUP);
    if (!gr) c->disable();

    new FXMenuSeparator(&pmenu);

    c=new FXMenuCommand(&pmenu, I18N("Edit Transfer"), NULL, this, ID_BUTTON_EDITTRANSFER);
    if (!st) c->disable();

    c=new FXMenuCommand(&pmenu, I18N("Delete Transfer"), NULL, this, ID_BUTTON_DELTRANSFER);
    if (!st) c->disable();

    new FXMenuSeparator(&pmenu);
    switch(m_tableType) {
    case AE_Book_TableType_BankTransferPlanned:
      c=new FXMenuCommand(&pmenu, I18N("Execute transfer"), NULL,
                          this, ID_BUTTON_EXECUTE);
      if (!st) c->disable();
      c=new FXMenuCommand(&pmenu, I18N("Move to \"Templates\" tab"), NULL,
                          this, ID_BUTTON_MOVETO_TEMPLATES);
      if (!st) c->disable();
      c=new FXMenuCommand(&pmenu, I18N("Copy to \"Templates\" tab"), NULL,
                          this, ID_BUTTON_COPYTO_TEMPLATES);
      if (!st) c->disable();
      break;

    case AE_Book_TableType_BankTransferTemplate:
      c=new FXMenuCommand(&pmenu, I18N("Execute transfer"), NULL,
                          this, ID_BUTTON_EXECUTE);
      if (!st) c->disable();
      c=new FXMenuCommand(&pmenu, I18N("Move to \"Planned\" tab"), NULL,
                          this, ID_BUTTON_MOVETO_PLANNED);
      if (!st) c->disable();
      c=new FXMenuCommand(&pmenu, I18N("Copy to \"Planned\" tab"), NULL,
                          this, ID_BUTTON_COPYTO_PLANNED);
      if (!st) c->disable();
      break;

    case AE_Book_TableType_BankTransfer:
      c=new FXMenuCommand(&pmenu, I18N("Execute transfer"), NULL,
                          this, ID_BUTTON_EXECUTE);
      if (!st) c->disable();
      c=new FXMenuCommand(&pmenu, I18N("Move to \"Templates\" tab"), NULL,
                          this, ID_BUTTON_MOVETO_TEMPLATES);
      if (!st) c->disable();
      c=new FXMenuCommand(&pmenu, I18N("Move to \"Planned\" tab"), NULL,
                          this, ID_BUTTON_MOVETO_PLANNED);
      if (!st) c->disable();
      c=new FXMenuCommand(&pmenu, I18N("Copy to \"Templates\" tab"), NULL,
                          this, ID_BUTTON_COPYTO_TEMPLATES);
      if (!st) c->disable();
      c=new FXMenuCommand(&pmenu, I18N("Copy to \"Planned\" tab"), NULL,
                          this, ID_BUTTON_COPYTO_PLANNED);
      if (!st) c->disable();

      c=new FXMenuCommand(&pmenu, I18N("Automatically reconcile"), NULL,
			  this, ID_BUTTON_AUTORECON);
      break;

    default:
      break;
    }

    new FXMenuSeparator(&pmenu);
    new FXMenuCommand(&pmenu, I18N("Back"), NULL, &pmenu, ID_UNPOST);

    pmenu.create();
    pmenu.popup(NULL, event->root_x, event->root_y);
    getApp()->runModalWhileShown(&pmenu);
  }

#endif
  return 1;
}



void FM_InvoicePage::dbOpened(AE_BOOK *b) {
  if (m_isVisible) {
    rebuild();
    m_needRebuild=false;
  }
  else {
    m_needRebuild=true;
  }
}



void FM_InvoicePage::dbClosing() {
  m_invoiceTree->clearItems();
}



void FM_InvoicePage::dbClosed() {
}



void FM_InvoicePage::dbReload(AE_BOOK_TABLE_TYPE tt) {
  if (tt==AE_Book_TableType_Job ||
      tt==AE_Book_TableType_None) {
    if (m_isVisible) {
      rebuild();
      m_needRebuild=false;
    }
    else
      m_needRebuild=true;
  }
}



void FM_InvoicePage::rebuild() {
  FF_App *app;
  AE_BOOK *b;
  AE_JOB *j;
  AQDB_ID currentInvId=0;

  j=m_invoiceTree->getCurrentInvoice();
  if (j)
    currentInvId=AE_Job_GetId(j);
  m_invoiceTree->clearItems();
  m_invoiceTree->reload();
  app=dynamic_cast<FF_App*>(getApp());
  b=app->getBook();
  if (b) {
    AE_JOB_LIST *jl;
    int rv;
    int i;

    jl=AE_Job_List_new();
    rv=AE_Book_GetOpenBills(b, m_isIncoming?1:0, jl);
    if (rv<0) {
      DBG_ERROR(0, "Error getting open bills (%d)", rv);
    }
    else {
      AE_JOB *j;

      /* add open bills to invoice tree */
      while( (j=AE_Job_List_First(jl)) ) {
	AE_Job_List_Del(j);
	m_invoiceTree->addInvoice(j);
      }
    }
    AE_Job_List_free(jl);

    /* sort list */
    for (i=0; i<m_invoiceTree->getNumHeaders(); i++) {
      FXbool b;
  
      b=m_invoiceTree->getHeaderArrowDir(i);
      if (b!=MAYBE) {
        if (b==TRUE)
          m_invoiceTree->sortByColumn(i, true);
        else
          m_invoiceTree->sortByColumn(i, false);
        break;
      }
    }
    m_invoiceTree->expandItems();
  }
  else {
    DBG_ERROR(0, "No book (%s)", m_isIncoming?"incoming":"outgoing");
  }

  if (currentInvId!=0) {
    FF_InvoiceTreeItem *item=m_invoiceTree->findInvoiceItem(currentInvId);
    if (item) {
      item->setSelected(true);
      m_invoiceTree->makeItemVisible(item);
      m_invoiceTree->setCurrentItem(item);
    }
    else {
      DBG_ERROR(0, "Item not found");
    }
  }
}



void FM_InvoicePage::enter() {
  m_isVisible=true;

  if (m_needRebuild) {
    rebuild();
    m_needRebuild=false;
  }
}



void FM_InvoicePage::leave() {
  m_isVisible=false;
}



void FM_InvoicePage::reset() {
  m_invoiceTree->clearItems();
}







