/****************************************************************************
 * This file is part of the project AqFinance.
 * AqFinance (c) by 2007 Martin Preuss, all rights reserved.
 *
 * The license for this file can be found in the file COPYING which you
 * should have received along with this file.
 ****************************************************************************/

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif


#include "ff_bankaccountlist.hpp"

#include <gwenhywfar/debug.h>
#include <gwenhywfar/i18n.h>


#define I18N(msg) GWEN_I18N_Translate(PACKAGE, msg)




class FF_BankAccountListItem: public FXListItem {
  FF_BankAccountListItem():FXListItem() {};
public:
  FF_BankAccountListItem(AE_BANKACCOUNT *ba,
                         FXIcon* icon=NULL);
  ~FF_BankAccountListItem();

  AE_BANKACCOUNT *getBankAccount() const;

};





FF_BankAccountListItem::FF_BankAccountListItem(AE_BANKACCOUNT *ba, FXIcon* icon)
:FXListItem("", icon) {
  FXString t;
  const char *s;

  s=AE_BankAccount_GetBankCode(ba);
  if (s)
    t+=s;
  t+="\t";

  s=AE_BankAccount_GetAccountNumber(ba);
  if (s)
    t+=s;
  t+="\t";

  s=AE_BankAccount_GetBic(ba);
  if (s)
    t+=s;
  t+="\t";

  s=AE_BankAccount_GetIban(ba);
  if (s)
    t+=s;
  t+="\t";

  setText(t);
  setData((void*) ba);
}



FF_BankAccountListItem::~FF_BankAccountListItem() {
  AE_BANKACCOUNT *ba;

  ba=(AE_BANKACCOUNT*) getData();
  if (ba) {
    AE_BankAccount_free(ba);
    setData(NULL);
  }
}



AE_BANKACCOUNT *FF_BankAccountListItem::getBankAccount() const {
  return (AE_BANKACCOUNT*) getData();

}






FF_BankAccountList::FF_BankAccountList()
:FXList() {

}



FF_BankAccountList::FF_BankAccountList(FXComposite *p, FXObject* tgt, FXSelector sel, FXuint opts,
                                       FXint x, FXint y, FXint w, FXint h)
:FXList(p, tgt, sel, opts, x, y, w, h) {
}



FF_BankAccountList::~FF_BankAccountList() {
}



void FF_BankAccountList::addBankAccount(AE_BANKACCOUNT *ba) {
  FF_BankAccountListItem *item;

  item=new FF_BankAccountListItem(ba, NULL);
  appendItem(item);
}



void FF_BankAccountList::addBankAccounts(AE_BANKACCOUNT_LIST2 *bl, int copyMode) {
  AE_BANKACCOUNT_LIST2_ITERATOR *it;

  it=AE_BankAccount_List2_First(bl);
  if (it) {
    AE_BANKACCOUNT *ba;
    FF_BankAccountListItem *item;

    ba=AE_BankAccount_List2Iterator_Data(it);

    switch(copyMode) {
      case COPYMODE_COPY:
        while(ba) {
          item=new FF_BankAccountListItem(AE_BankAccount_dup(ba), NULL);
          appendItem(item);
          ba=AE_BankAccount_List2Iterator_Next(it);
        }
        break;
  
      case COPYMODE_ATTACH:
        while(ba) {
          AE_BankAccount_Attach(ba);
          item=new FF_BankAccountListItem(ba, NULL);
          appendItem(item);
          ba=AE_BankAccount_List2Iterator_Next(it);
        }
        break;
      case COPYMODE_NONE:
      default:
        while(ba) {
          item=new FF_BankAccountListItem(ba, NULL);
          appendItem(item);
          ba=AE_BankAccount_List2Iterator_Next(it);
        }
        break;
    }
    AE_BankAccount_List2Iterator_free(it);
  }
}



void FF_BankAccountList::removeBankAccount(AQDB_ID id) {
  FXint n;
  FXint i;

  n=getNumItems();
  for (i=0; i<n; i++) {
    FXListItem *item=getItem(i);
    if (item) {
      FF_BankAccountListItem *baItem=dynamic_cast<FF_BankAccountListItem*>(item);
      if (baItem) {
        AE_BANKACCOUNT *ba=baItem->getBankAccount();
        if (AE_BankAccount_GetId(ba)==id) {
          removeItem(i);
          delete baItem;
          break;
        }
      }
    }
  }
}



void FF_BankAccountList::getSelectedBankAccounts(AE_BANKACCOUNT_LIST2 *bl, int copyMode) {
  FXint n;
  FXint i;

  n=getNumItems();

  switch(copyMode) {
    case COPYMODE_COPY:
      for (i=0; i<n; i++) {
        FXListItem *item=getItem(i);
        if (item && item->isSelected()) {
          FF_BankAccountListItem *baItem=dynamic_cast<FF_BankAccountListItem*>(item);
          if (baItem) {
            AE_BANKACCOUNT *ba;
  
            ba=baItem->getBankAccount();
            AE_BankAccount_dup(ba);
            AE_BankAccount_List2_PushBack(bl, AE_BankAccount_dup(ba));
          }
        }
      }
      break;
  
    case COPYMODE_ATTACH:
      for (i=0; i<n; i++) {
        FXListItem *item=getItem(i);
        if (item && item->isSelected()) {
          FF_BankAccountListItem *baItem=dynamic_cast<FF_BankAccountListItem*>(item);
          if (baItem) {
            AE_BANKACCOUNT *ba;
  
            ba=baItem->getBankAccount();
            AE_BankAccount_Attach(ba);
            AE_BankAccount_List2_PushBack(bl, ba);
          }
        }
      }
      break;
  
    case COPYMODE_NONE:
    default:
      for (i=0; i<n; i++) {
        FXListItem *item=getItem(i);
        if (item && item->isSelected()) {
          FF_BankAccountListItem *baItem=dynamic_cast<FF_BankAccountListItem*>(item);
          if (baItem)
            AE_BankAccount_List2_PushBack(bl, baItem->getBankAccount());
        }
      }
      break;
  }
}



AE_BANKACCOUNT *FF_BankAccountList::getCurrentBankAccount() {
  FXint cit=getCurrentItem();

  if (cit>=0) {
    FXListItem *item=getItem(cit);
    if (item) {
      FF_BankAccountListItem *baItem=dynamic_cast<FF_BankAccountListItem*>(item);
      if (baItem)
        return baItem->getBankAccount();
    }
  }

  return NULL;
}



AE_BANKACCOUNT *FF_BankAccountList::getFirstSelectedBankAccount() {
  FXint n;
  FXint i;

  n=getNumItems();

  for (i=0; i<n; i++) {
    FXListItem *item=getItem(i);
    if (item && item->isSelected()) {
      FF_BankAccountListItem *baItem=dynamic_cast<FF_BankAccountListItem*>(item);
      if (baItem)
        return baItem->getBankAccount();
    }
  }

  return NULL;
}



void FF_BankAccountList::selectBankAccount(AQDB_ID id) {
  FXint n;
  FXint i;

  n=getNumItems();
  for (i=0; i<n; i++) {
    FXListItem *item=getItem(i);
    if (item) {
      FF_BankAccountListItem *baItem=dynamic_cast<FF_BankAccountListItem*>(item);
      if (baItem) {
        AE_BANKACCOUNT *ba=baItem->getBankAccount();
        if (AE_BankAccount_GetId(ba)==id) {
          selectItem(i, false);
          break;
        }
      }
    }
  }
}



void FF_BankAccountList::deselectBankAccount(AQDB_ID id) {
  FXint n;
  FXint i;

  n=getNumItems();
  for (i=0; i<n; i++) {
    FXListItem *item=getItem(i);
    if (item) {
      FF_BankAccountListItem *baItem=dynamic_cast<FF_BankAccountListItem*>(item);
      if (baItem) {
        AE_BANKACCOUNT *ba=baItem->getBankAccount();
        if (AE_BankAccount_GetId(ba)==id) {
          deselectItem(i, false);
          break;
        }
      }
    }
  }
}






