/****************************************************************************
 * This file is part of the project AqFinance.
 * AqFinance (c) by 2010 Martin Preuss, all rights reserved.
 *
 * The license for this file can be found in the file COPYING which you
 * should have received along with this file.
 ****************************************************************************/

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif



#include "ff_invoicetree.hpp"
#include "ff_app.hpp"

#include <fx.h>

#include <aqfinance/engine/book/ae_job.h>
#include <aqfinance/engine/book/ae_utils.h>

#include <gwenhywfar/debug.h>
#include <gwenhywfar/i18n.h>

#define I18N(msg) GWEN_I18N_Translate(PACKAGE, msg)




FF_InvoiceTreeItem::FF_InvoiceTreeItem(const FXString &txt,
				       AE_JOB *j,
				       FXIcon* oi, FXIcon* ci, void* ptr)
:FF_GroupTreeItem(oi, ci, ptr)
,m_job(j)
{
  setText(txt);
}



FF_InvoiceTreeItem::~FF_InvoiceTreeItem() {
  if (m_isDataItem && m_job) {
    AE_Job_free(m_job);
  }
}









FF_InvoiceTree::FF_InvoiceTree(FXComposite *p,
			       FXObject* tgt, FXSelector sel,
			       FXuint opts,
			       FXint x, FXint y, FXint w, FXint h)
:FF_GroupTree(p, tgt, sel, opts, x, y, w, h)
,m_partners(NULL)
,m_commodities(NULL)
{
  int idx=0;

  appendHeader(I18N("Priority"));
  setHeaderSize(idx++, 100);
  appendHeader(I18N("Enter Date"));
  setHeaderSize(idx++, 100);
  appendHeader(I18N("Due Date"));
  setHeaderSize(idx++, 100);
  appendHeader(I18N("Partner"));
  setHeaderSize(idx++, 100);
  appendHeader(I18N("Memo"));
  setHeaderSize(idx++, 100);
  appendHeader(I18N("Amount"));
  setHeaderSize(idx++, 100);
}



FF_InvoiceTree::~FF_InvoiceTree() {
  AE_Partner_List_free(m_partners);
  AE_Commodity_List_free(m_commodities);
}



FF_InvoiceTreeItem *FF_InvoiceTree::findInvoiceItem(AQDB_ID id) {
  FXFoldingItem *item;

  item=getFirstItem();
  while (item) {
    FF_GroupTreeItem *grItem;

    grItem=dynamic_cast<FF_GroupTreeItem*>(item);
    if (grItem) {
      FF_InvoiceTreeItem *res=findInvoiceItem(grItem, id);
      if (res!=NULL)
        return res;
    }
    item=item->getNext();
  }

  return NULL;
}



FF_InvoiceTreeItem *FF_InvoiceTree::findInvoiceItem(FF_GroupTreeItem *grItem,
						    AQDB_ID id) {
  if (grItem->isDataItem()) {
    FF_InvoiceTreeItem *jItem=dynamic_cast<FF_InvoiceTreeItem*>(grItem);
    if (jItem) {
      AE_JOB *j=jItem->getJob();
      if (AE_Job_GetId(j)==id)
        return jItem;
    }
  }
  else {
    FXFoldingItem *item;

    item=grItem->getFirst();
    while(item) {
      grItem=dynamic_cast<FF_GroupTreeItem*>(item);
      if (grItem) {
        FF_InvoiceTreeItem *res=findInvoiceItem(grItem, id);
        if (res!=NULL)
          return res;
      }
      item=item->getNext();
    }
  }

  return NULL;
}



FF_InvoiceTreeItem *FF_InvoiceTree::findInvoiceItem(const AE_JOB *qj) {
  FXFoldingItem *item;

  item=getFirstItem();
  while (item) {
    FF_GroupTreeItem *grItem=dynamic_cast<FF_GroupTreeItem*>(item);
    if (grItem && (grItem->isDataItem())) {
      FF_InvoiceTreeItem *jItem;

      jItem=dynamic_cast<FF_InvoiceTreeItem*>(grItem);
      if (jItem) {
        AE_JOB *j;

	j=jItem->getJob();
	if (j==qj)
          return jItem;
      }
    }
    item=item->getBelow();
  }

  return NULL;
}



void FF_InvoiceTree::addInvoice(AE_JOB *j) {
  FF_GroupTreeItem *itemFather=NULL;
  AQDB_ID id;
  FF_InvoiceTreeItem *item;
  FXString txt;
  GWEN_BUFFER *tbuf;
  int rv;
  const GWEN_DATE *dt;
  const char *s;

#if 0
  id=AE_Job_GetGroupId(j);
  if (id) {
    itemFather=findGroupItem(id);
    if (itemFather==NULL) {
      DBG_ERROR(AE_LOGDOMAIN, "Group for job not found, reassigning to root");
    }
  }
  else {
    DBG_ERROR(0, "Job has no gid");
  }
#endif

  // TODO: create text
  tbuf=GWEN_Buffer_new(0, 64, 0, 1);

  /* priority */
  txt+=FXStringVal(AE_Job_GetPriority(j), 10);
  txt+="\t";

  /* enter date */
  dt=AE_Job_GetEnterDate(j);
  if (dt) {
    /* translators: the following is a template for the local date */
    GWEN_Date_toStringWithTemplate(dt, I18N("YYYY/MM/DD"), tbuf);
    txt+=FXString(GWEN_Buffer_GetStart(tbuf));
    GWEN_Buffer_Reset(tbuf);
  }
  txt+="\t";

  /* due date */
  dt=AE_Job_GetDueDate(j);
  if (dt) {
    /* translators: the following is a template for the local date */
    GWEN_Date_toStringWithTemplate(dt, I18N("YYYY/MM/DD"), tbuf);
    txt+=FXString(GWEN_Buffer_GetStart(tbuf));
    GWEN_Buffer_Reset(tbuf);
  }
  txt+="\t";

  /* partner */
  id=AE_Job_GetPartnerId(j);
  if (id>0) {
    AE_PARTNER *pt;

    pt=AE_Partner_List_GetById(m_partners, id);
    if (pt) {
      AE_Partner_PathToBuffer(pt, tbuf);
      txt+=FXString(GWEN_Buffer_GetStart(tbuf));
      GWEN_Buffer_Reset(tbuf);
    }
  }
  txt+="\t";

  s=AE_Job_GetMemo(j);
  if (s && *s)
    txt+=FXString(s);
  txt+="\t";

  /* amount */
  rv=AE_Utils_AmountToString(AE_Job_GetTotalSum(j),
			     AE_Job_GetCommodityId(j),
			     m_commodities,
			     tbuf);
  if (rv==0)
    txt+=FXString(GWEN_Buffer_GetStart(tbuf));

  /* done */
  GWEN_Buffer_free(tbuf);
  item=new FF_InvoiceTreeItem(txt, j,
			      /*m_iconOpen, m_iconClosed*/
			      NULL, NULL); /* no icons here */
  appendItem(itemFather, item, false);
}



void FF_InvoiceTree::addInvoice(const AE_JOB *j) {
  addInvoice(AE_Job_dup(j));
}



FF_InvoiceTree::FF_InvoiceTree(): FF_GroupTree() {
}



AE_JOB *FF_InvoiceTree::getCurrentInvoice() {
  FXFoldingItem *item;

  item=getCurrentItem();
  if (item) {
    FF_GroupTreeItem *grItem=dynamic_cast<FF_GroupTreeItem*>(item);
    if (grItem && (grItem->isDataItem())) {
      FF_InvoiceTreeItem *jItem;

      jItem=dynamic_cast<FF_InvoiceTreeItem*>(grItem);
      if (jItem)
        return jItem->getJob();
    }
  }

  return NULL;
}



AE_JOB *FF_InvoiceTree::getFirstSelectedJob() {
  FXFoldingItem *item;

  item=getFirstItem();
  while (item) {
    if (item->isSelected()) {
      FF_GroupTreeItem *grItem=dynamic_cast<FF_GroupTreeItem*>(item);
      if (grItem && (grItem->isDataItem())) {
        FF_InvoiceTreeItem *jItem;

        jItem=dynamic_cast<FF_InvoiceTreeItem*>(grItem);
        if (jItem)
          return jItem->getJob();
      }
    }
    item=item->getBelow();
  }

  return NULL;
}



void FF_InvoiceTree::getSelectedInvoices(AE_JOB_LIST2 *al) {
  FXFoldingItem *item;

  item=getFirstItem();
  while (item) {
    if (item->isSelected()) {
      FF_GroupTreeItem *grItem=dynamic_cast<FF_GroupTreeItem*>(item);
      if (grItem && (grItem->isDataItem())) {
        FF_InvoiceTreeItem *jItem=dynamic_cast<FF_InvoiceTreeItem*>(grItem);
        if (jItem)
          AE_Job_List2_PushBack(al, jItem->getJob());
      }
    }
    item=item->getBelow();
  }
}



void FF_InvoiceTree::selectInvoice(AQDB_ID id, bool b) {
  FXFoldingItem *item;

  item=getFirstItem();
  while (item) {
    FF_GroupTreeItem *grItem=dynamic_cast<FF_GroupTreeItem*>(item);
    if (grItem && (grItem->isDataItem())) {
      FF_InvoiceTreeItem *jItem;

      jItem=dynamic_cast<FF_InvoiceTreeItem*>(grItem);
      if (jItem) {
        AE_JOB *j;

        j=jItem->getJob();
        if (AE_Job_GetId(j)==id) {
          jItem->setSelected(b);
          break;
        }
      }
    }
    item=item->getBelow();
  }
}



void FF_InvoiceTree::selectInvoice(const AE_JOB *j, bool b) {
  FXFoldingItem *item;

  item=getFirstItem();
  while (item) {
    FF_GroupTreeItem *grItem=dynamic_cast<FF_GroupTreeItem*>(item);
    if (grItem && (grItem->isDataItem())) {
      FF_InvoiceTreeItem *jItem;

      jItem=dynamic_cast<FF_InvoiceTreeItem*>(grItem);
      if (jItem) {
	const AE_JOB *ij;

        ij=jItem->getJob();
        if (ij==j) {
	  jItem->setSelected(b);
          break;
        }
      }
    }
    item=item->getBelow();
  }
}



void FF_InvoiceTree::getInvoices(AE_JOB_LIST2 *al, int copyMode) {
  FXFoldingItem *item;

  switch(copyMode) {
  case COPYMODE_NONE:
    item=getFirstItem();
    while (item) {
      FF_GroupTreeItem *grItem=dynamic_cast<FF_GroupTreeItem*>(item);
      if (grItem && (grItem->isDataItem())) {
	FF_InvoiceTreeItem *jItem=dynamic_cast<FF_InvoiceTreeItem*>(grItem);
	if (jItem)
	  AE_Job_List2_PushBack(al, jItem->getJob());
      }
      item=item->getBelow();
    }
    break;

  case COPYMODE_COPY:
    item=getFirstItem();
    while (item) {
      FF_GroupTreeItem *grItem=dynamic_cast<FF_GroupTreeItem*>(item);
      if (grItem && (grItem->isDataItem())) {
	FF_InvoiceTreeItem *jItem=dynamic_cast<FF_InvoiceTreeItem*>(grItem);
	if (jItem) {
	  AE_JOB *j;

	  j=jItem->getJob();
	  assert(j);
	  AE_Job_List2_PushBack(al, AE_Job_dup(j));
	}
      }
      item=item->getBelow();
    }
    break;

  default:
  case COPYMODE_ATTACH:
    item=getFirstItem();
    while (item) {
      FF_GroupTreeItem *grItem=dynamic_cast<FF_GroupTreeItem*>(item);
      if (grItem && (grItem->isDataItem())) {
	FF_InvoiceTreeItem *jItem=dynamic_cast<FF_InvoiceTreeItem*>(grItem);
	if (jItem) {
	  AE_JOB *j;

	  j=jItem->getJob();
	  assert(j);
	  AE_Job_Attach(j);
	  AE_Job_List2_PushBack(al, j);
	}
      }
      item=item->getBelow();
    }
    break;
  }
}



void FF_InvoiceTree::makeInvoiceVisible(AQDB_ID id) {
  FXFoldingItem *item;

  item=getFirstItem();
  while (item) {
    FF_GroupTreeItem *grItem=dynamic_cast<FF_GroupTreeItem*>(item);
    if (grItem && (grItem->isDataItem())) {
      FF_InvoiceTreeItem *jItem;

      jItem=dynamic_cast<FF_InvoiceTreeItem*>(grItem);
      if (jItem) {
        AE_JOB *j;

        j=jItem->getJob();
        if (AE_Job_GetId(j)==id) {
          makeItemVisible(jItem);
          break;
        }
      }
    }
    item=item->getBelow();
  }
}




#define DEF_SORT_FN_UP_TXT(macro_num, macro_name)                                            \
FXint FF_InvoiceTree::sortFunctionUp##macro_num(const FXFoldingItem *it1,                    \
                                                const FXFoldingItem *it2) {                  \
  const FF_GroupTreeItem *grItem1=dynamic_cast<const FF_GroupTreeItem*>(it1);                \
  const FF_GroupTreeItem *grItem2=dynamic_cast<const FF_GroupTreeItem*>(it2);                \
                                                                                             \
  if (grItem1 && grItem2) {                                                                  \
    if (!grItem1->isDataItem() && !grItem2->isDataItem())                                    \
      return FF_GroupTree::sortFunctionUp(it1, it2);                                         \
    else if (!grItem1->isDataItem())                                                         \
      return 1;                                                                              \
    else if (!grItem2->isDataItem())                                                         \
      return -1;                                                                             \
    else {                                                                                   \
      const FF_InvoiceTreeItem *tti1=dynamic_cast<const FF_InvoiceTreeItem*>(it1);           \
      const FF_InvoiceTreeItem *tti2=dynamic_cast<const FF_InvoiceTreeItem*>(it2);           \
      const AE_JOB *ba1=tti1->getJob();                                                      \
      const AE_JOB *ba2=tti2->getJob();                                                      \
      const char *s1=AE_Job_Get##macro_name(ba1);                                            \
      const char *s2=AE_Job_Get##macro_name(ba2);                                            \
                                                                                             \
      /* both are data items */                                                              \
      if (s1 && s2)                                                                          \
        return strcasecmp(s1, s2);                                                           \
      else if (s1)                                                                           \
        return 1;                                                                            \
      else if (s2)                                                                           \
        return -1;                                                                           \
    }                                                                                        \
  }                                                                                          \
                                                                                             \
  return 0;                                                                                  \
}



#define DEF_SORT_FN_UP_INT(macro_num, macro_name)                                            \
FXint FF_InvoiceTree::sortFunctionUp##macro_num(const FXFoldingItem *it1,                    \
                                                const FXFoldingItem *it2) {                  \
  const FF_GroupTreeItem *grItem1=dynamic_cast<const FF_GroupTreeItem*>(it1);                \
  const FF_GroupTreeItem *grItem2=dynamic_cast<const FF_GroupTreeItem*>(it2);                \
                                                                                             \
  if (grItem1 && grItem2) {                                                                  \
    if (!grItem1->isDataItem() && !grItem2->isDataItem())                                    \
      return FF_GroupTree::sortFunctionUp(it1, it2);                                         \
    else if (!grItem1->isDataItem())                                                         \
      return 1;                                                                              \
    else if (!grItem2->isDataItem())                                                         \
      return -1;                                                                             \
    else {                                                                                   \
      const FF_InvoiceTreeItem *tti1=dynamic_cast<const FF_InvoiceTreeItem*>(it1);           \
      const FF_InvoiceTreeItem *tti2=dynamic_cast<const FF_InvoiceTreeItem*>(it2);           \
      const AE_JOB *ba1=tti1->getJob();                                                      \
      const AE_JOB *ba2=tti2->getJob();                                                      \
      int s1=AE_Job_Get##macro_name(ba1);                                                    \
      int s2=AE_Job_Get##macro_name(ba2);                                                    \
                                                                                             \
      /* both are data items */                                                              \
      if (s1==s2)                                                                            \
        return 0;                                                                            \
      else if (s1>s2)                                                                        \
        return 1;                                                                            \
      else                                                                                   \
        return -1;                                                                           \
    }                                                                                        \
  }                                                                                          \
                                                                                             \
  return 0;                                                                                  \
}


#define DEF_SORT_FN_UP_DATE(macro_num, macro_name)                                           \
FXint FF_InvoiceTree::sortFunctionUp##macro_num(const FXFoldingItem *it1,                    \
                                                const FXFoldingItem *it2) {                  \
  const FF_GroupTreeItem *grItem1=dynamic_cast<const FF_GroupTreeItem*>(it1);                \
  const FF_GroupTreeItem *grItem2=dynamic_cast<const FF_GroupTreeItem*>(it2);                \
                                                                                             \
  if (grItem1 && grItem2) {                                                                  \
    if (!grItem1->isDataItem() && !grItem2->isDataItem())                                    \
      return FF_GroupTree::sortFunctionUp(it1, it2);                                         \
    else if (!grItem1->isDataItem())                                                         \
      return 1;                                                                              \
    else if (!grItem2->isDataItem())                                                         \
      return -1;                                                                             \
    else {                                                                                   \
      const FF_InvoiceTreeItem *tti1=dynamic_cast<const FF_InvoiceTreeItem*>(it1);           \
      const FF_InvoiceTreeItem *tti2=dynamic_cast<const FF_InvoiceTreeItem*>(it2);           \
      const AE_JOB *ba1=tti1->getJob();                                                      \
      const AE_JOB *ba2=tti2->getJob();                                                      \
      const GWEN_DATE *s1=AE_Job_Get##macro_name(ba1);                                       \
      const GWEN_DATE *s2=AE_Job_Get##macro_name(ba2);                                       \
                                                                                             \
      /* both are data items */                                                              \
      if (s1 && s2)                                                                          \
        return GWEN_Date_Compare(s1, s2);                                                    \
      else if (s1)                                                                           \
        return 1;                                                                            \
      else if (s2)                                                                           \
        return -1;                                                                           \
    }                                                                                        \
  }                                                                                          \
                                                                                             \
  return 0;                                                                                  \
}



#define DEF_SORT_FN_UP_VALUE(macro_num, macro_name)                                          \
FXint FF_InvoiceTree::sortFunctionUp##macro_num(const FXFoldingItem *it1,                    \
                                                const FXFoldingItem *it2) {                  \
  const FF_GroupTreeItem *grItem1=dynamic_cast<const FF_GroupTreeItem*>(it1);                \
  const FF_GroupTreeItem *grItem2=dynamic_cast<const FF_GroupTreeItem*>(it2);                \
                                                                                             \
  if (grItem1 && grItem2) {                                                                  \
    if (!grItem1->isDataItem() && !grItem2->isDataItem())                                    \
      return FF_GroupTree::sortFunctionUp(it1, it2);                                         \
    else if (!grItem1->isDataItem())                                                         \
      return 1;                                                                              \
    else if (!grItem2->isDataItem())                                                         \
      return -1;                                                                             \
    else {                                                                                   \
      const FF_InvoiceTreeItem *tti1=dynamic_cast<const FF_InvoiceTreeItem*>(it1);           \
      const FF_InvoiceTreeItem *tti2=dynamic_cast<const FF_InvoiceTreeItem*>(it2);           \
      const AE_JOB *ba1=tti1->getJob();                                                      \
      const AE_JOB *ba2=tti2->getJob();                                                      \
      const AQDB_VALUE *s1=AE_Job_Get##macro_name(ba1);                                      \
      const AQDB_VALUE *s2=AE_Job_Get##macro_name(ba2);                                      \
                                                                                             \
      /* both are data items */                                                              \
      if (s1 && s2)                                                                          \
        return AQDB_Value_Compare(s1, s2);                                                   \
      else if (s1)                                                                           \
        return 1;                                                                            \
      else if (s2)                                                                           \
        return -1;                                                                           \
    }                                                                                        \
  }                                                                                          \
                                                                                             \
  return 0;                                                                                  \
}


#define DEF_SORT_FN_DOWN(macro_num)                                                          \
FXint FF_InvoiceTree::sortFunctionDown##macro_num(const FXFoldingItem *it1,                  \
                                                  const FXFoldingItem *it2){                 \
  FXint rv;                                                                                  \
                                                                                             \
  rv=sortFunctionUp##macro_num(it1, it2);                                                    \
  if (rv<0)                                                                                  \
    return 1;                                                                                \
  else if (rv>0)                                                                             \
    return -1;                                                                               \
  return 0;                                                                                  \
}






DEF_SORT_FN_UP_INT(0, Priority)
DEF_SORT_FN_DOWN(0)
DEF_SORT_FN_UP_DATE(1, EnterDate)
DEF_SORT_FN_DOWN(1)
DEF_SORT_FN_UP_DATE(2, DueDate)
DEF_SORT_FN_DOWN(2)
DEF_SORT_FN_UP_TXT(4, Memo)
DEF_SORT_FN_DOWN(4)
DEF_SORT_FN_UP_VALUE(5, TotalSum)
DEF_SORT_FN_DOWN(5)




#undef DEF_SORT_FN_UP_VALUE
#undef DEF_SORT_FN_UP_DATE
#undef DEF_SORT_FN_UP_INT
#undef DEF_SORT_FN_UP_TXT
#undef DEF_SORT_FN_DOWN




void FF_InvoiceTree::sortByColumn(int c, bool up) {
  switch(c) {
  case 0:
    setSortFunc(up?sortFunctionUp0:sortFunctionDown0);
    sortItems();
    break;
  case 1:
    setSortFunc(up?sortFunctionUp1:sortFunctionDown1);
    sortItems();
    break;
  case 2:
    setSortFunc(up?sortFunctionUp2:sortFunctionDown2);
    sortItems();
    break;
  case 4:
    setSortFunc(up?sortFunctionUp4:sortFunctionDown4);
    sortItems();
    break;
  case 5:
    setSortFunc(up?sortFunctionUp5:sortFunctionDown5);
    sortItems();
    break;
  default:
    DBG_ERROR(0, "Sort by column %d (%s) not implemented", c, up?"up":"down");
    break;
  }
}



void FF_InvoiceTree::reload() {
  FF_App *app;
  AE_BOOK *b;
  
  app=dynamic_cast<FF_App*>(getApp());
  b=app->getBook();
  if (b) {
    AE_PARTNER_LIST *pList;
    AE_COMMODITY_LIST *cList;
    int rv;

    /* read parters */
    pList=AE_Partner_List_new();
    rv=AE_Book_ReadPartnerList(b, AE_Book_TableType_Partner, pList);
    if (rv<0) {
      DBG_INFO(AE_LOGDOMAIN, "here (%d)", rv);
      AE_Partner_List_free(pList);
    }
    else {
      AE_Partner_List_free(m_partners);
      m_partners=pList;
    }

    /* read commodities */
    cList=AE_Commodity_List_new();
    rv=AE_Book_ReadCommodityList(b, AE_Book_TableType_Commodity, cList);
    if (rv<0) {
      DBG_INFO(AE_LOGDOMAIN, "here (%d)", rv);
      AE_Commodity_List_free(cList);
    }
    else {
      AE_Commodity_List_free(m_commodities);
      m_commodities=cList;
    }
  }
}







