/****************************************************************************
 * This file is part of the project AqFinance.
 * AqFinance (c) by 2013 Martin Preuss, all rights reserved.
 *
 * The license for this file can be found in the file COPYING which you
 * should have received along with this file.
 ****************************************************************************/

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif



#include "ff_sharetree.hpp"

#include <fx.h>


#include <aqfinance/engine/book/ae_share.h>


#include <gwenhywfar/debug.h>
#include <gwenhywfar/i18n.h>
#include <gwenhywfar/gwentime.h>


#define I18N(msg) GWEN_I18N_Translate(PACKAGE, msg)





FF_ShareTreeItem::FF_ShareTreeItem(AE_SHARE *st, void* ptr)
:FF_GroupTreeItem(NULL, NULL, ptr)
,m_share(st)
{
}



FF_ShareTreeItem::~FF_ShareTreeItem() {
  if (m_isDataItem && m_share) {
    AE_Share_free(m_share);
    m_share=(AE_SHARE*) -1;
  }
}




FXString FF_ShareTreeItem::getItemText(const FXXCheckTree* list) const {
  const FF_ShareTree *tree;
  FXString t;

  tree=dynamic_cast<const FF_ShareTree*>(list);
  if (tree) {
    int i;

    for (i=0; i<tree->m_columns; i++) {
      const char *s;

      switch(tree->m_columnTypes[i]) {
      case FF_ShareTree::COLUMN_NAME:
	s=AE_Share_GetName(m_share);
        if (s)
          t+=FXString(s);
        break;

      case FF_ShareTree::COLUMN_ISIN:
	s=AE_Share_GetIsin(m_share);
        if (s)
          t+=FXString(s);
        break;

      case FF_ShareTree::COLUMN_DOWNLOADNAME:
	s=AE_Share_GetDownloadName(m_share);
        if (s)
          t+=FXString(s);
	break;

      case FF_ShareTree::COLUMN_AMOUNT:
        {
          const AQDB_VALUE *v;

          v=AE_Share_GetAmount(m_share);
          if (v) {
            GWEN_BUFFER *tbuf=GWEN_Buffer_new(0, 256, 0, 1);

            t+="<right>";

            AQDB_Value_toHumanReadableString(v, tbuf, 2);
            t+=GWEN_Buffer_GetStart(tbuf);
	    GWEN_Buffer_free(tbuf);
	    // TODO: Get commodity
	    t+="</right>";
          }
        }
        break;

      case FF_ShareTree::COLUMN_NOTE:
	s=AE_Share_GetNote(m_share);
	if (s)
	  t+=FXString(s);
	break;

      case FF_ShareTree::COLUMN_LOWERLIMIT:
        {
          const AQDB_VALUE *v;

          v=AE_Share_GetLowerLimit(m_share);
          if (v) {
            GWEN_BUFFER *tbuf=GWEN_Buffer_new(0, 256, 0, 1);

            t+="<right>";

            AQDB_Value_toHumanReadableString(v, tbuf, 2);
            t+=GWEN_Buffer_GetStart(tbuf);
	    GWEN_Buffer_free(tbuf);
	    // TODO: Get commodity
	    t+="</right>";
          }
        }
        break;

      case FF_ShareTree::COLUMN_UPPERLIMIT:
        {
          const AQDB_VALUE *v;

          v=AE_Share_GetUpperLimit(m_share);
          if (v) {
            GWEN_BUFFER *tbuf=GWEN_Buffer_new(0, 256, 0, 1);

            t+="<right>";

            AQDB_Value_toHumanReadableString(v, tbuf, 2);
            t+=GWEN_Buffer_GetStart(tbuf);
	    GWEN_Buffer_free(tbuf);
	    // TODO: Get commodity
	    t+="</right>";
          }
        }
        break;

      case FF_ShareTree::COLUMN_NEXT_STOCKHOLDER_MEETING:
        {
          const GWEN_DATE *ti;

          ti=AE_Share_GetNextStockholdersMeeting(m_share);
          if (ti) {
            GWEN_BUFFER *tbuf=GWEN_Buffer_new(0, 256, 0, 1);
            GWEN_Date_toStringWithTemplate(ti, I18N("YYYY/MM/DD"), tbuf);
            t+=GWEN_Buffer_GetStart(tbuf);
            GWEN_Buffer_free(tbuf);
          }
        }
        break;

      case FF_ShareTree::COLUMN_LAST_STOCKHOLDER_MEETING:
        {
          const GWEN_DATE *ti;

          ti=AE_Share_GetLastStockholdersMeeting(m_share);
          if (ti) {
            GWEN_BUFFER *tbuf=GWEN_Buffer_new(0, 256, 0, 1);
            GWEN_Date_toStringWithTemplate(ti, I18N("YYYY/MM/DD"), tbuf);
            t+=GWEN_Buffer_GetStart(tbuf);
            GWEN_Buffer_free(tbuf);
          }
        }
        break;

      case FF_ShareTree::COLUMN_NEXT_REPORT_DATE:
        {
          const GWEN_DATE *ti;

          ti=AE_Share_GetNextReportDate(m_share);
          if (ti) {
            GWEN_BUFFER *tbuf=GWEN_Buffer_new(0, 256, 0, 1);
            GWEN_Date_toStringWithTemplate(ti, I18N("YYYY/MM/DD"), tbuf);
            t+=GWEN_Buffer_GetStart(tbuf);
            GWEN_Buffer_free(tbuf);
          }
        }
        break;

      case FF_ShareTree::COLUMN_DIVIDEND:
        {
          const AQDB_VALUE *v;

          v=AE_Share_GetDividend(m_share);
	  if (v) {
	    GWEN_BUFFER *tbuf=GWEN_Buffer_new(0, 256, 0, 1);

	    t+="<right>";

            AQDB_Value_toHumanReadableString(v, tbuf, 3);
            t+=GWEN_Buffer_GetStart(tbuf);
	    GWEN_Buffer_free(tbuf);
	    t+="</right>";
	  }
        }
        break;

      default:
	break;
      }

      t+="\t";
    } /* for */
  }

  return t;
}



FXString FF_ShareTreeItem::getColumnTooltipText(FXint col) {
  return "";
}






FF_ShareTree::FF_ShareTree(FXComposite *p, FXObject* tgt, FXSelector sel, FXuint opts,
			   FXint x, FXint y, FXint w, FXint h)
:FF_GroupTree(p, tgt, sel, opts, x, y, w, h)
,m_columns(0)
{
}



FF_ShareTree::FF_ShareTree()
:FF_GroupTree() {
}



FF_ShareTree::~FF_ShareTree() {
}



void FF_ShareTree::create() {
  FF_GroupTree::create();
}



FF_ShareTreeItem *FF_ShareTree::findShareItem(AQDB_ID id) {
  FXFoldingItem *item;

  item=getFirstItem();
  while (item) {
    FF_GroupTreeItem *grItem;

    grItem=dynamic_cast<FF_GroupTreeItem*>(item);
    if (grItem) {
      FF_ShareTreeItem *res=findShareItem(grItem, id);
      if (res!=NULL)
        return res;
    }
    item=item->getNext();
  }

  return NULL;
}



FF_ShareTreeItem *FF_ShareTree::findShareItem(FF_GroupTreeItem *grItem, AQDB_ID id) {
  if (grItem->isDataItem()) {
    FF_ShareTreeItem *stItem=dynamic_cast<FF_ShareTreeItem*>(grItem);
    if (stItem) {
      AE_SHARE *st=stItem->getShare();
      if (AE_Share_GetId(st)==id)
	return stItem;
    }
  }
  else {
    FXFoldingItem *item;

    item=grItem->getFirst();
    while(item) {
      grItem=dynamic_cast<FF_GroupTreeItem*>(item);
      if (grItem) {
        FF_ShareTreeItem *res=findShareItem(grItem, id);
	if (res!=NULL)
	  return res;
      }
      item=item->getNext();
    }
  }

  return NULL;
}



FF_ShareTreeItem *FF_ShareTree::findShareItem(const AE_SHARE *sh) {
  FXFoldingItem *item;

  item=getFirstItem();
  while (item) {
    FF_GroupTreeItem *grItem=dynamic_cast<FF_GroupTreeItem*>(item);
    if (grItem && (grItem->isDataItem())) {
      FF_ShareTreeItem *baItem;

      baItem=dynamic_cast<FF_ShareTreeItem*>(grItem);
      if (baItem) {
	AE_SHARE *sta;

	sta=baItem->getShare();
	if (sta==sh)
	  return baItem;
      }
    }
    item=item->getBelow();
  }

  return NULL;
}




void FF_ShareTree::addShare(AE_SHARE *st, bool direct, int useColor) {
  FF_GroupTreeItem *itemFather=NULL;

  if (!direct) {
    AQDB_ID gid=AE_Share_GetGroupId(st);
    if (gid) {
      itemFather=findGroupItem(gid);
      if (itemFather==NULL) {
        DBG_INFO(AE_LOGDOMAIN, "Group for share not found, reassigning to root");
      }
    }
  }

  addShareToItem(st, itemFather, useColor);
}



void FF_ShareTree::addShares(AE_SHARE_LIST2 *sl, int copyMode, bool direct) {
  AE_SHARE_LIST2_ITERATOR *it;

  it=AE_Share_List2_First(sl);
  if (it) {
    AE_SHARE *st;

    st=AE_Share_List2Iterator_Data(it);
    switch(copyMode) {
    case COPYMODE_COPY:
      while(st) {
        addShare(AE_Share_dup(st), direct, USECOLOR_NORMAL);
        st=AE_Share_List2Iterator_Next(it);
      }
      break;

    case COPYMODE_ATTACH:
      while(st) {
        AE_Share_Attach(st);
        addShare(st, direct, USECOLOR_NORMAL);
        st=AE_Share_List2Iterator_Next(it);
      }
      break;

    case COPYMODE_NONE:
    default:
      while(st) {
        addShare(st, direct, USECOLOR_NORMAL);
        st=AE_Share_List2Iterator_Next(it);
      }
      break;
    }
    AE_Share_List2Iterator_free(it);
  }
}



void FF_ShareTree::addShareToGroup(AE_SHARE *st, AQDB_ID gid, int useColor) {
  FF_GroupTreeItem *itemFather=NULL;

  if (gid) {
    itemFather=findGroupItem(gid);
    if (itemFather==NULL) {
      DBG_INFO(AE_LOGDOMAIN, "Group for share not found, reassigning to root");
    }
  }

  addShareToItem(st, itemFather, useColor);
}



void FF_ShareTree::addShareToItem(AE_SHARE *st, FF_GroupTreeItem *itemFather, int useColor) {
  FF_ShareTreeItem *item;

  if (itemFather==NULL) {
    DBG_ERROR(0, "No fatherItem");
  }

  item=new FF_ShareTreeItem(st);
  /* maybe colorize */
  switch(useColor) {
  case FF_ShareTree::USECOLOR_NOTICE:
    item->setUseOwnColors(true);
    item->setBackColor(FXRGB(247, 248, 96));
    item->setTextColor(FXRGB(0, 0, 0));
    break;
  case FF_ShareTree::USECOLOR_WARNING:
    item->setUseOwnColors(true);
    item->setBackColor(FXRGB(255, 127, 0));
    item->setTextColor(FXRGB(255, 255, 255));
    break;

  case FF_ShareTree::USECOLOR_NORMAL:
  default:
    item->setUseOwnColors(false);
    break;
  }

  appendItem(itemFather, item, false);
}



AE_SHARE *FF_ShareTree::getCurrentShare() {
  FXFoldingItem *item;

  item=getCurrentItem();
  if (item) {
    FF_GroupTreeItem *grItem=dynamic_cast<FF_GroupTreeItem*>(item);
    if (grItem && (grItem->isDataItem())) {
      FF_ShareTreeItem *stItem;

      stItem=dynamic_cast<FF_ShareTreeItem*>(grItem);
      if (stItem)
	return stItem->getShare();
    }
  }

  return NULL;
}



AE_SHARE *FF_ShareTree::getFirstSelectedShare() {
  FXFoldingItem *item;

  item=getFirstItem();
  while (item) {
    if (item->isSelected()) {
      FF_GroupTreeItem *grItem=dynamic_cast<FF_GroupTreeItem*>(item);
      if (grItem && (grItem->isDataItem())) {
        FF_ShareTreeItem *stItem;

        stItem=dynamic_cast<FF_ShareTreeItem*>(grItem);
        if (stItem)
          return stItem->getShare();
      }
    }
    item=item->getBelow();
  }

  return NULL;
}



void FF_ShareTree::getSelectedShares(AE_SHARE_LIST2 *sl, int copyMode) {
  FXFoldingItem *item;

  switch(copyMode) {
  case COPYMODE_COPY:
    item=getFirstItem();
    while (item) {
      if (item->isSelected()) {
        FF_GroupTreeItem *grItem=dynamic_cast<FF_GroupTreeItem*>(item);
        if (grItem && (grItem->isDataItem())) {
          FF_ShareTreeItem *stItem=dynamic_cast<FF_ShareTreeItem*>(grItem);
          if (stItem)
	    AE_Share_List2_PushBack(sl, AE_Share_dup(stItem->getShare()));
        }
      }
      item=item->getBelow();
    }
    break;

  case COPYMODE_ATTACH:
    item=getFirstItem();
    while (item) {
      if (item->isSelected()) {
        FF_GroupTreeItem *grItem=dynamic_cast<FF_GroupTreeItem*>(item);
        if (grItem && (grItem->isDataItem())) {
          FF_ShareTreeItem *stItem=dynamic_cast<FF_ShareTreeItem*>(grItem);
	  if (stItem) {
	    AE_SHARE *st=stItem->getShare();
            AE_Share_Attach(st);
            AE_Share_List2_PushBack(sl, st);
          }
        }
      }
      item=item->getBelow();
    }
    break;

  case COPYMODE_NONE:
  default:
    item=getFirstItem();
    while (item) {
      if (item->isSelected()) {
        FF_GroupTreeItem *grItem=dynamic_cast<FF_GroupTreeItem*>(item);
        if (grItem && (grItem->isDataItem())) {
          FF_ShareTreeItem *stItem=dynamic_cast<FF_ShareTreeItem*>(grItem);
          if (stItem)
	    AE_Share_List2_PushBack(sl, stItem->getShare());
	}
      }
      item=item->getBelow();
    }
    break;
  }
}



void FF_ShareTree::selectShare(AQDB_ID id, bool b) {
  FXFoldingItem *item;

  item=getFirstItem();
  while (item) {
    FF_GroupTreeItem *grItem=dynamic_cast<FF_GroupTreeItem*>(item);
    if (grItem && (grItem->isDataItem())) {
      FF_ShareTreeItem *stItem;

      stItem=dynamic_cast<FF_ShareTreeItem*>(grItem);
      if (stItem) {
	AE_SHARE *st;

        st=stItem->getShare();
	if (AE_Share_GetId(st)==id) {
	  stItem->setSelected(b);
	  break;
	}
      }
    }
    item=item->getBelow();
  }
}



void FF_ShareTree::selectShare(const AE_SHARE *st, bool b) {
  FXFoldingItem *item;

  item=getFirstItem();
  while (item) {
    FF_GroupTreeItem *grItem=dynamic_cast<FF_GroupTreeItem*>(item);
    if (grItem && (grItem->isDataItem())) {
      FF_ShareTreeItem *stItem;

      stItem=dynamic_cast<FF_ShareTreeItem*>(grItem);
      if (stItem) {
	if (stItem->getShare()==st) {
          stItem->setSelected(b);
          break;
        }
      }
    }
    item=item->getBelow();
  }
}



void FF_ShareTree::makeShareVisible(AQDB_ID id) {
  FXFoldingItem *item;

  item=getFirstItem();
  while (item) {
    FF_GroupTreeItem *grItem=dynamic_cast<FF_GroupTreeItem*>(item);
    if (grItem && (grItem->isDataItem())) {
      FF_ShareTreeItem *stItem;

      stItem=dynamic_cast<FF_ShareTreeItem*>(grItem);
      if (stItem) {
        AE_SHARE *st;

        st=stItem->getShare();
	if (AE_Share_GetId(st)==id) {
          makeItemVisible(stItem);
          break;
        }
      }
    }
    item=item->getBelow();
  }
}



void FF_ShareTree::addColumn(int c, int w) {
  if (m_columns<COLUMN_MAX) {
    switch(c) {
    case COLUMN_NAME:
      appendHeader(I18N("ShareTree|Name"));
      setHeaderSize(m_columns, w);
      m_columnTypes[m_columns++]=c;
      break;
    case COLUMN_ISIN:
      appendHeader(I18N("ShareTree|ISIN"));
      setHeaderSize(m_columns, w);
      m_columnTypes[m_columns++]=c;
      break;
    case COLUMN_DOWNLOADNAME:
      appendHeader(I18N("ShareTree|Download Name"));
      setHeaderSize(m_columns, w);
      m_columnTypes[m_columns++]=c;
      break;
    case COLUMN_AMOUNT:
      appendHeader(I18N("ShareTree|Amount"));
      setHeaderSize(m_columns, w);
      m_columnTypes[m_columns++]=c;
      break;
    case COLUMN_NOTE:
      appendHeader(I18N("ShareTree|Note"));
      setHeaderSize(m_columns, w);
      m_columnTypes[m_columns++]=c;
      break;
    case COLUMN_LOWERLIMIT:
      appendHeader(I18N("ShareTree|Lower Limit"));
      setHeaderSize(m_columns, w);
      m_columnTypes[m_columns++]=c;
      break;
    case COLUMN_UPPERLIMIT:
      appendHeader(I18N("ShareTree|Upper Limit"));
      setHeaderSize(m_columns, w);
      m_columnTypes[m_columns++]=c;
      break;
    case COLUMN_NEXT_STOCKHOLDER_MEETING:
      appendHeader(I18N("ShareTree|Next Shareholder Meeting"));
      setHeaderSize(m_columns, w);
      m_columnTypes[m_columns++]=c;
      break;
    case COLUMN_LAST_STOCKHOLDER_MEETING:
      appendHeader(I18N("ShareTree|Lat Shareholder Meeting"));
      setHeaderSize(m_columns, w);
      m_columnTypes[m_columns++]=c;
      break;
    case COLUMN_NEXT_REPORT_DATE:
      appendHeader(I18N("ShareTree|Next Report Date"));
      setHeaderSize(m_columns, w);
      m_columnTypes[m_columns++]=c;
      break;
    case COLUMN_DIVIDEND:
      appendHeader(I18N("ShareTree|Dividend"));
      setHeaderSize(m_columns, w);
      m_columnTypes[m_columns++]=c;
      break;
    default:
      break;
    }
  }
}



FXX_CHECKTREE_SORT_FN_UP_TXT(FF_ShareTree, FF_ShareTreeItem, getShare, AE_SHARE, AE_Share, Name);
FXX_CHECKTREE_SORT_FN_DOWN(FF_ShareTree, FF_ShareTreeItem, getShare, AE_SHARE, AE_Share, Name);

FXX_CHECKTREE_SORT_FN_UP_TXT(FF_ShareTree, FF_ShareTreeItem, getShare, AE_SHARE, AE_Share, Isin);
FXX_CHECKTREE_SORT_FN_DOWN(FF_ShareTree, FF_ShareTreeItem, getShare, AE_SHARE, AE_Share, Isin);

FXX_CHECKTREE_SORT_FN_UP_TXT(FF_ShareTree, FF_ShareTreeItem, getShare, AE_SHARE, AE_Share, DownloadName);
FXX_CHECKTREE_SORT_FN_DOWN(FF_ShareTree, FF_ShareTreeItem, getShare, AE_SHARE, AE_Share, DownloadName);

FXX_CHECKTREE_SORT_FN_UP_VALUE(FF_ShareTree, FF_ShareTreeItem, getShare, AE_SHARE, AE_Share, Amount);
FXX_CHECKTREE_SORT_FN_DOWN(FF_ShareTree, FF_ShareTreeItem, getShare, AE_SHARE, AE_Share, Amount);

FXX_CHECKTREE_SORT_FN_UP_TXT(FF_ShareTree, FF_ShareTreeItem, getShare, AE_SHARE, AE_Share, Note);
FXX_CHECKTREE_SORT_FN_DOWN(FF_ShareTree, FF_ShareTreeItem, getShare, AE_SHARE, AE_Share, Note);

FXX_CHECKTREE_SORT_FN_UP_VALUE(FF_ShareTree, FF_ShareTreeItem, getShare, AE_SHARE, AE_Share, LowerLimit);
FXX_CHECKTREE_SORT_FN_DOWN(FF_ShareTree, FF_ShareTreeItem, getShare, AE_SHARE, AE_Share, LowerLimit);

FXX_CHECKTREE_SORT_FN_UP_VALUE(FF_ShareTree, FF_ShareTreeItem, getShare, AE_SHARE, AE_Share, UpperLimit);
FXX_CHECKTREE_SORT_FN_DOWN(FF_ShareTree, FF_ShareTreeItem, getShare, AE_SHARE, AE_Share, UpperLimit);

FXX_CHECKTREE_SORT_FN_UP_DATE(FF_ShareTree, FF_ShareTreeItem, getShare, AE_SHARE, AE_Share, NextStockholdersMeeting);
FXX_CHECKTREE_SORT_FN_DOWN(FF_ShareTree, FF_ShareTreeItem, getShare, AE_SHARE, AE_Share, NextStockholdersMeeting);

FXX_CHECKTREE_SORT_FN_UP_DATE(FF_ShareTree, FF_ShareTreeItem, getShare, AE_SHARE, AE_Share, LastStockholdersMeeting);
FXX_CHECKTREE_SORT_FN_DOWN(FF_ShareTree, FF_ShareTreeItem, getShare, AE_SHARE, AE_Share, LastStockholdersMeeting);

FXX_CHECKTREE_SORT_FN_UP_DATE(FF_ShareTree, FF_ShareTreeItem, getShare, AE_SHARE, AE_Share, NextReportDate);
FXX_CHECKTREE_SORT_FN_DOWN(FF_ShareTree, FF_ShareTreeItem, getShare, AE_SHARE, AE_Share, NextReportDate);

FXX_CHECKTREE_SORT_FN_UP_VALUE(FF_ShareTree, FF_ShareTreeItem, getShare, AE_SHARE, AE_Share, Dividend);
FXX_CHECKTREE_SORT_FN_DOWN(FF_ShareTree, FF_ShareTreeItem, getShare, AE_SHARE, AE_Share, Dividend);




void FF_ShareTree::sortByColumn(int c, bool up) {
  if (c<m_columns) {
    switch(m_columnTypes[c]) {
    case COLUMN_NAME:
      setSortFunc(up?sortFunctionUpName:sortFunctionDownName);
      sortItems();
      break;
    case COLUMN_ISIN:
      setSortFunc(up?sortFunctionUpIsin:sortFunctionDownIsin);
      sortItems();
      break;
    case COLUMN_DOWNLOADNAME:
      setSortFunc(up?sortFunctionUpDownloadName:sortFunctionDownDownloadName);
      sortItems();
      break;
    case COLUMN_AMOUNT:
    case COLUMN_NOTE:
      setSortFunc(up?sortFunctionUpNote:sortFunctionDownNote);
      sortItems();
      break;
    case COLUMN_LOWERLIMIT:
      setSortFunc(up?sortFunctionUpLowerLimit:sortFunctionDownLowerLimit);
      sortItems();
      break;
    case COLUMN_UPPERLIMIT:
      setSortFunc(up?sortFunctionUpUpperLimit:sortFunctionDownUpperLimit);
      sortItems();
      break;
    case COLUMN_NEXT_STOCKHOLDER_MEETING:
      setSortFunc(up?sortFunctionUpNextStockholdersMeeting:sortFunctionDownNextStockholdersMeeting);
      sortItems();
      break;
    case COLUMN_LAST_STOCKHOLDER_MEETING:
      setSortFunc(up?sortFunctionUpLastStockholdersMeeting:sortFunctionDownLastStockholdersMeeting);
      sortItems();
      break;
    case COLUMN_NEXT_REPORT_DATE:
      setSortFunc(up?sortFunctionUpNextReportDate:sortFunctionDownNextReportDate);
      sortItems();
      break;
    case COLUMN_DIVIDEND:
      setSortFunc(up?sortFunctionUpDividend:sortFunctionDownDividend);
      sortItems();
      break;
    default:
      DBG_ERROR(0, "Sort by column %d (%s) not implemented", c, up?"up":"down");
      break;
    }
  }
}





